import pandas as pd
import numpy
import pathlib
import json
import datetime

## script that runs on package import to check format of config file, existance of csvs, and columns ##
def check_struc(config):
    """
    Checks to ensure the config structure is valid
    """
    ## config requirements ##
    config_requiremnets = {
        'name' : {'type':str, 'nullable' : False},
        'description' : {'type':str, 'nullable' : False},
        'last_local_update' : {'type':str, 'nullable' : True},
        'download_url' : {'type':str, 'nullable' : False},
        'compression' : {'type':str, 'nullable' : True},
        'engine' : {'type':str, 'nullable' : True},
        'iter' : {'type':dict, 'nullable' : False},
        'assignments' : {'type':list, 'nullable' : False},
        'map' : {'type':dict, 'nullable' : False},
    }
    ## containers ##
    passing = True
    errors = []
    for k, v in config_requiremnets.items():
        if k not in config:
            ## check that config has the prop ##
            passing = False
            errors.append({
                'type' : 'config_structure',
                'error_msg' : '{0} is a required property, but was not found in the map config'.format(
                    k
                )
            })
        else:
            ## make sure its of the right type ##
            if config[k] is None and not v['nullable']:
                passing = False
                errors.append({
                    'type' : 'config_structure',
                    'error_msg' : '{0} is cannot be null'.format(
                        k
                    )
                })
            elif config[k] is not None:
                if not isinstance(config[k], v['type']):
                    errors.append({
                        'type' : 'config_structure',
                        'error_msg' : '{0} must be a {1}'.format(
                            k, v['type']
                        )
                    })
    ## return ##            
    return passing, errors

def check_map_type(map):
    """
    Iterates through map keys value pairs and ensures values are an allowed type
    """
    allowed_types = [
        'object', 'int32', 'int64', 'float32', 'float64',
        'boolean', 'interval', 'category', 'datetime64[ns, <tz>]',
        'bool'
    ]
    ## containers ##
    passing = True
    errors = []
    for k, v in map.items():
        if v not in allowed_types:
            passing = False
            errors.append({
                'error_type' : 'dtype',
                'error_msg' : '{0} type provided for {1} is not an accepted datatype'.format(
                    v, k 
                )
            })
    ## return ##
    return passing, errors




import requests
import datetime
import pathlib
import json

class Freshness():
    """
    Determines whether or not a data update is required
    """
    
    def __init__(self, config):
        self.config = config
        self.needs_update = False
        self.freshness_check_time = None
        self.check_freshness()
        
    ## helpers ##  
    def determine_current_season(self):
        """
        Loads the current season from global variables
        """
        ## container for current seaosn ##
        season = None
        ## load global_variables.json ##
        with open('{0}/Utilities/global_variables.json'.format(pathlib.Path(__file__).parent.parent.parent.resolve()), 'r') as fp:
            global_variables = json.load(fp)
            season = global_variables['season_states']['last_full_week']['season']
        ## return ##
        return season
        
    def form_season_iter_download_url(self, download_url):
        """
        If the dataset is a season iter, formulate the dl link based on current seaosn
        get current season
        """
        year = self.determine_current_season()
        if year is None:
            raise ValueError('ERROR: Could not determine current season.')
        return download_url.format(year)
    
    ## primary SLA check ##
    def sla_check(self):
        """
        Checks SLA seconds and the last update timestamp from the config, to determine if
        a freshness check is needed, or if the data is still within its SLA
        """
        ## fails SLA if null ##
        if self.config['freshness']['last_freshness_check'] is None or self.config['freshness']['sla_seconds'] is None:
            ## either freshness was never previously established or there is no SLA and freshness should be checked ##
            return False
        else:
            ## get seconds since last freshness check ##
            last_freshness_check = datetime.datetime.fromisoformat(
                self.config['freshness']['last_freshness_check']
            ).astimezone(datetime.timezone.utc)
            seconds_since_last_check = (
                datetime.datetime.now(datetime.timezone.utc) -
                last_freshness_check
            ).total_seconds()
            ## compare to sla ##
            if seconds_since_last_check > self.config['freshness']['sla_seconds']:
                ## if freshness state breaches SLA, update freshness
                return False
            else:
                ## if the time since the last check is less than the SLA, then it passes
                return True
    
    ## config types ##
    def check_gh_release_freshness(self):
        """
        Checks the freshness of a map with a github release source.
        """
        ## helpers ##
        def get_gh_release_date(github_endpoint, tag, public_download_url):
            ## Retrieves releases for a tag, matches last update on dl url, and returns
            ## the timestampe of the last update.
            ## get release ##
            r = requests.get(
                '{0}/{1}'.format(
                    github_endpoint,
                    tag
                )
            )
            ## hand response ##
            if r.status_code != 200:
                print('WARN: Request to github releases failed for {0}.'.format(tag))
                return
            ## load json ##
            r = r.json()
            ## get assets ##
            assets = r.get('assets')
            if assets is None:
                print('WARN: Release assets did not come with response json from github for {0}.'.format(tag))
                return
            if len(assets) == 0:
                print('WARN: No release assets included in response json from github for {0}.'.format(tag))
                return
            ## iter through assets until url matches ##
            last_update = None
            for asset in assets:
                if asset['browser_download_url'] == public_download_url:
                    try:
                        return datetime.datetime.fromisoformat(
                            asset['updated_at']
                        ).astimezone(datetime.timezone.utc)
                    except:
                        print('WARN: Could not parse last update string for {0}.'.format(public_download_url))
                        return
            ## if nothing has been returned, return None ##
            print('WARN: No asset matched {0}.'.format(public_download_url))
        
        def check_for_new_gh_data(last_local_update, github_endpoint, tag, public_download_url):
            ## Compares the last update of the local file to the last update of the
            ## release asset on github to determine if new data is available.
            ## get last update ##
            last_gh_update = get_gh_release_date(github_endpoint, tag, public_download_url)
            ## compare ##
            if last_gh_update is None:
                return False
            if last_local_update is None:
                ## if no local update, return True signal download ##
                return True
            if last_gh_update > datetime.datetime.fromisoformat(
                last_local_update
            ).astimezone(datetime.timezone.utc):
                ## if gh update is newer, return True signal download ##
                return True
            return False
            
        ## handle request for GH Release freshness ##
        github_endpoint = self.config['freshness'].get('gh_api_endpoint')
        tag = self.config['freshness'].get('gh_release_tag')
        download_url = self.config.get('download_url')
        if tag is None or download_url is None or github_endpoint is None:
            raise ValueError('ERROR: Tag, Endpoint, and Download URL must be included in a Github Release Map')
        ## get url of most recent season file if iter ##
        if self.config['iter']['type'] is not None:
            if self.config['iter']['type'] == 'season':
                download_url = self.form_season_iter_download_url(download_url)
        last_local_update = self.config.get('last_local_update')
        ## check last GH update ##
        self.needs_update = check_for_new_gh_data(
            last_local_update,
            github_endpoint,
            tag,
            download_url
        )
        
    def check_gh_commit_freshness(self):
        """
        Checks the freshness of a map with a github commit source.
        """
        ## helpers ##
        def get_gh_commit_date(endpoint):
            ## retrieves the last time a commit was pushed to a repo ##
            ## note, this does not necessarily mean the data itself was updated, but ##
            ## its the safest way to ensure fresh data w/o a release ##
            r = requests.get(endpoint)
            ## handle response ##
            if r.status_code != 200:
                print('WARN: Request to github commits failed.')
                return
            r = r.json()
            if len(r) == 0:
                print('WARN: No commits returned from github.')
                return
            ## parse results ##
            commits = []
            for commit in r:
                commits.append(
                    datetime.datetime.fromisoformat(
                        commit['commit']['committer']['date']
                    ).astimezone(datetime.timezone.utc)
                )
            ## return ##
            return max(commits)
        
        def check_for_new_gh_data(last_local_update, endpoint):
            ## Compares the last update of the local file to the last update of the
            ## commit on github to determine if new data is available.
            ## get last update ##
            last_gh_update = get_gh_commit_date(endpoint)
            ## compare ##
            if last_gh_update is None:
                return False
            if last_local_update is None:
                ## if no local update, return True signal download ##
                return True
            if last_gh_update > datetime.datetime.fromisoformat(
                last_local_update
            ).astimezone(datetime.timezone.utc):
                ## if gh update is newer, return True signal download ##
                return True
            return False
        
        ## handle request for GH Commit freshness ##
        github_endpoint = self.config['freshness'].get('gh_api_endpoint')
        if github_endpoint is None:
            raise ValueError('ERROR: Endpoint must be included in a Github Commit Map')
        last_local_update = self.config.get('last_local_update')
        ## check last GH update ##
        self.needs_update = check_for_new_gh_data(
            last_local_update,
            github_endpoint
        )
    
    def check_freshness(self):
        """
        Wrapper function that checks data freshness for a map given a freshness config.
        """
        ## check the freshness sla ##
        sla_passing = self.sla_check()
        ## update if it is not passing ##
        if not sla_passing:
            ## check the data type and route to appropriate function ##
            if self.config['freshness']['type'] == 'gh_release':
                self.check_gh_release_freshness()
            elif self.config['freshness']['type'] == 'gh_commit':
                self.check_gh_commit_freshness()
            else:
                raise ValueError('ERROR: Freshness type not recognized.')
            ## update last check property with the current time since freshness was checked ##
            self.freshness_check_time = datetime.datetime.now(datetime.timezone.utc).isoformat()



import pandas as pd
import numpy
import pathlib
import datetime

from .Primatives import TableMap, Freshness, DataPull, LocalIO
import nfelodcm.nfelodcm.Utilities as utils

class DCMTable():
    """
    Takes a table name and handles all backend loading and pulling of data.
    """
    def __init__(self, table):
        self.table = table
        ## init table map ##
        self.tableMap = TableMap(self.table)
        self.check_table_map()
        ## init freshness ##
        self.freshness = Freshness(self.tableMap.config)
        ## init data pull ##
        self.dataPull = DataPull(self.tableMap.config)
        ## init local io ##
        self.localIO = LocalIO(self.tableMap.config, self.tableMap.table_data_location)
        ## handle load ##
        self.handle_load()

    def check_table_map(self):
        """
        Checks to ensure the structure of the table map config and data types 
        are valid
        """
        ##
        config_passes, config_errors = utils.check_struc(self.tableMap.config)
        if not config_passes:
            raise ValueError(
                "Config structure is not valid. Errors: \n     -> " +
                '\n     -> '.join([err['error_msg'] for err in config_errors])
            )
        ## check map ##
        map_passes, map_errors = utils.check_map_type(self.tableMap.config['map'])
        if not map_passes:
            raise ValueError(
                "Map had data type errors. Errors: \n     -> " +
                '\n     -> '.join([err['error_msg'] for err in map_errors])
            )


    def handle_load(self):
        """
        Handles loading of the table based on freshness status.
        """
        if self.freshness.freshness_check_time is not None:
            ## if freshness was checked, update the time in the config ##
            self.tableMap.config['freshness']['last_freshness_check'] = self.freshness.freshness_check_time
            self.tableMap.update_map(
                self.tableMap.config
            )
        ## reference freshness to determine where to pull data from ##
        if self.freshness.needs_update:
            ## pull new data ##
            self.dataPull.update_data()
            ## load into local io ##
            self.localIO.df = self.dataPull.pulled_df
            ## write ##
            self.localIO.write()
            ## update config timestamp with a UTC iso timestamp ##
            self.tableMap.config['last_local_update'] = datetime.datetime.utcnow().isoformat()
            self.tableMap.update_map(
                self.tableMap.config
            )
        else:
            self.localIO.read()



import pandas as pd
import numpy

from .Engine import DCMTable
import nfelodcm.nfelodcm.Utilities as utils

## init season state on load ##
utils.set_season_state()

## wrappers and classes to interact with the package ##
def get_df(table):
    """
    gets a single table
    """
    engine = DCMTable(table)
    return engine.localIO.df

def load(tables):
    """
    loads an array of tables and returns a dictionary
    """
    ## init db structure ##
    db = {}
    ## add tables ##
    for table in tables:
        db[table] = get_df(table)
    ## return db struc ##
    return db

def get_map(url):
    '''
    reads a web based csv and returns a column map
    with 64 to 32 downcasting
    '''
    df = pd.read_csv(url)
    for col in df.columns.tolist():
        dtype = df[col].dtype.name
        ## downcast ##
        if '64' in dtype:
            dtype = dtype.replace('64', '32')
        ## print ##
        print('"{0}": "{1}",'.format(
            col, dtype
        ))
        


import pandas as pd

## assignment functions that replaces fastr team names with legacy team names ##
repl = {
    'LV' : 'OAK',
    'SD' : 'LAC',
    'STL' : 'LAR',
    'LA' : 'LAR',
}

def game_id_repl(df):
    """
    Replaces fastr game_ids with a legacy game id.
    """
    df['game_id'] = (
        df['season'].astype('str') + '_' +
        df['week'].astype('str').str.zfill(2) + '_' +
        df['away_team'].replace(repl) + '_' +
        df['home_team'].replace(repl)
    )
    return df
